package kosoft.dbgen.generate;

import java.util.*;
import java.util.Date;
import java.io.*;

import java.sql.*;

import kosoft.dbgen.util.*;


public class BuildClassesV2 {
    
    ArrayList tableList = new ArrayList();
    
    String prefix = "au/com/cytopia/lims/db";
    String pkgprefix = "au.com.cytopia.lims.db";
    String implPrefix = "au/com/cytopia/lims/dbimpl";
    String implPkgPrefix = "au.com.cytopia.lims.dbimpl";
    String dbConnectionName="pepcatdb";

    RuntimeProperties properties = new RuntimeProperties();
    
    void readParameters() throws Exception {
        properties.setFileName("jdbcgen.properties");
        properties.load();
        
        prefix = properties.getProperty("prefix", "au/com/cytopia/lims/db");
        pkgprefix = properties.getProperty("pkgPrefix", "au.com.cytopia.lims.db");
        implPrefix = properties.getProperty("implPrefix", "au/com/cytopia/lims/dbimpl");
        implPkgPrefix = properties.getProperty("implPkgPrefix", "au.com.cytopia.lims.dbimpl");
        dbConnectionName = properties.getProperty("dbConnectionName", "aConnection");

    }
    
    String getPathToJavaInterface() { return prefix + ""; }
    String getInterfacePackageName() { return pkgprefix + ""; }

    String getPathToJavaGenInterface() { return prefix + File.separator + "gen"; }
    String getGenInterfacePackageName() { return pkgprefix + ".gen"; }

/*    
    String getPathToJavaSource() { return prefix + "/data"; }
    String getJavaPackageName() { return pkgprefix + ".data"; }

    String getPathToJavaGenSource() { return prefix + "/data"; }
    String getGenJavaPackageName() { return pkgprefix + ".data"; }
 */
   
   String getSourcePathName(String mode) {
       String pathName = implPrefix + File.separator + mode.toLowerCase();
        File file = new File(pathName);
        file.mkdirs();
        return pathName;
    }
    
    String stripUnderscore(String mode) {
        String name = mode.toLowerCase();
        if (name.startsWith("_")) {
            name = name.substring(1);
            System.out.println("removing _" + ": " + name);
        }
        return name.toLowerCase();
    }

    String getExtensionForMode(String mode) {
        return "_" + mode + "Impl";
    }
    
    String getPackageName(String mode)  {
          return implPkgPrefix + "." + stripUnderscore(mode); 
    }    

    String getGenSourcePathName(String mode) { 
        String pathName = implPrefix + File.separator + stripUnderscore(mode) + File.separator + "gen";
        File file = new File(pathName);
        file.mkdirs();
        return pathName;
    }
    String getGenPackageName(String mode) { return implPkgPrefix + "." + stripUnderscore(mode) + ".gen"; }

    String getClientPathToJavaGenSource() { return implPrefix +"/client/gen"; }
    String getClientGenJavaPackageName() { return implPkgPrefix +".client.gen"; }

    String getClientPathToJavaSource() { return implPrefix + "/client"; }
    String getClientJavaPackageName() { return implPkgPrefix +".client"; }

    String getServerPathToJavaSource() { return implPrefix +"/server"; }
    String getServerJavaPackageName() { return implPkgPrefix +".server"; }

    String getDbConnectionName() { return dbConnectionName; }
    String getDbConnectionClassName() { return getDbConnectionName() + "Connection"; }
    
    

    void setupDirs() {
        
        (new File(getPathToJavaInterface())).mkdirs();
        (new File(getPathToJavaGenInterface())).mkdirs();

//        (new File(getPathToJavaSource())).mkdirs();
//        (new File(getPathToJavaGenSource())).mkdirs();

        (new File(getClientPathToJavaSource())).mkdirs();
        (new File(getClientPathToJavaGenSource())).mkdirs();

        (new File(getServerPathToJavaSource())).mkdirs();
    }
    
    String buildFileName(String prefix, String name) {
        String fileName = prefix + File.separator + name + ".java";
        return fileName;
    }

    String getBasicImportLines() {
        String result = "import java.util.*;\n";
        result += "import java.sql.*;\n";
        result +="import java.sql.Date;\n";
        return result;
    }

    
    void buildDBConnection() {
    
        String className = getDbConnectionClassName();
        String fileName  = buildFileName(getPathToJavaInterface(),className);

        // Check if file has already been generated
        
            File file = new File(fileName);
            
            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

        OutputFile out = new OutputFile(fileName);
        
        out.println("package " + getInterfacePackageName() + ";\n");

        out.println("import java.io.*;");
        out.println(getBasicImportLines());
        out.println("import kosoft.dbgen.util.*;");
        out.println("import " + getGenInterfacePackageName() + ".*;\n\n");

        out.println("public interface " +  className + " extends " + className + "Base {\n");
        out.println("\n		// Generated Class which user can extend\n\n");
            
        out.println("}");
        out.close();
    }
    
    void buildDBConnectionBase() {
    
        String className = getDbConnectionName() + "ConnectionBase";
        String fileName  = buildFileName(getPathToJavaGenInterface(),className);
        OutputFile out = new OutputFile(fileName);
        
        out.println("package " + getGenInterfacePackageName() + ";\n");
        out.println("import java.io.*;");
        out.println(getBasicImportLines());
        out.println("import kosoft.dbgen.util.*;");
        out.println("import " + getInterfacePackageName() + ".*;");
        out.println("");

        out.println("public interface " + className + " extends DbGenConnection {\n");
            
//        out.println("    public String getObjectTypeName() { return \"" + clientConnectionName +"\"; }\n");
    
    
    // Constructor

        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);
            out.println("    " + obj.getFactoryMethodPrototype() + ";");
        }
//        out.println("\n  }\n\n");        
        out.println("}");
        out.close();
    }

    
    void buildDbConnectionImpl(String mode) {
        String extension = getExtensionForMode(mode);
        String className = getDbConnectionClassName() + extension;
        String fileName  = buildFileName(getSourcePathName(mode),className);
        // Check if file has already been generated
        
        File file = new File(fileName);
        
        if (file.exists()) {
            System.out.println("file " + fileName + " already created");
            return;
        }

        OutputFile out = new OutputFile(fileName);
        
        out.println("package " + getPackageName(mode) + ";\n");
        out.println("import java.io.*;");
        out.println(getBasicImportLines());
        out.println("import kosoft.dbgen.util.*;");
        out.println("import kosoft.dbgen.rdb.*;");
        out.println("import " + getInterfacePackageName() + ".*;");
        out.println("import " + getGenPackageName(mode) + ".*;\n");
//        out.println("import kosoft.dbgen.client.gen.*;\n\n");

        out.println("public class " + className + " extends " + getDbConnectionClassName() + extension + "Base {\n");
        out.println("\n		// Generated Class which user can extend\n\n");
            
        out.println("}");
        out.close();
    }
    
    void buildInterfaces() {
        
        buildDBConnection();
        buildDBConnectionBase();
        
        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);
            obj.printFieldsInterface();
            obj.printObjectInterface();
            obj.printObjectInterfaceBase();
            
            obj.printObjectStmtInterfaceBase();
            obj.printObjectStmtInterface();
            
            obj.printObjectFactoryInterfaceBase();
            obj.printObjectFactoryInterface();
        }
        
    }
    
    void buildRdbImpl() {
        String mode = "Rdb";
        buildDbConnectionImplBase(mode);
        buildDbConnectionImpl(mode);
        
        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);            

            obj.printObjectStmtBase(mode);
            obj.printObjectStmt(mode);

            obj.printObjectBase(mode);
            obj.printObject(mode);
            
            obj.printObjectFactoryBase(mode);
            obj.printObjectFactory(mode);
            //  obj.printFieldsInterface();
        }

    }    

    void buildPython() {
        String mode = "Rdb";
        //buildDbConnectionImplBase(mode);
        //buildDbConnectionImpl(mode);

//        String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";
        String fileName = "fbassocdb.py";
        OutputFile out = new OutputFile(fileName);
        
        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);            

            obj.printPythonObjectBase(out);
            //  obj.printFieldsInterface();
        }

    }    
    
    void buildDbConnectionImplBase(String mode) {
    
        String extension = getExtensionForMode(mode);
        String className = getDbConnectionClassName() + extension + "Base";
        String fileName  = buildFileName(getGenSourcePathName(mode),className);
        OutputFile out = new OutputFile(fileName);
        
        out.println("package " + getGenPackageName(mode) + ";\n");
        out.println("import java.io.*;");
        out.println(getBasicImportLines());
        out.println("import kosoft.dbgen.util.*;");
        out.println("import kosoft.dbgen.rdb.*;");
        out.println("import " + getInterfacePackageName() + ".*;");
        out.println("import " + getPackageName(mode) + ".*;\n");
//        out.println("import kosoft.dbgen.client.gen.*;\n\n");

        out.println("public abstract class " + className + " extends RdbConnection implements " + getDbConnectionClassName() + " {\n");

        // Attributes 
        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);
            out.println("    " + obj.getFactoryName() + extension + " " + obj.getFactoryVariableName() + ";");
            //            out.println("    public" + obj.getFactoryName() + " " + obj.getFactoryVariableName() + ";");
        }
    
    // Constructor
    
        out.println("\n    public " + className + "() {");
        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);
            out.println("        " + obj.getFactoryVariableName() + " = new " + obj.getFactoryName() + extension + "();");
            out.println("        " + obj.getFactoryVariableName() + ".setRdbConnection(this);");
            out.println("");
        }
        
        out.println("    }\n");

        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);
            out.println("    " + obj.getFactoryMethodPrototype() + " {");
            out.println("        return " + obj.getFactoryVariableName() + ";");
            out.println("    }\n");
        }
        out.println("}");
        out.close();
    }
    
    void buildDbClientConnection() {
    
        String className = getDbConnectionName() + "ClientConnection";
        String fileName  = buildFileName(getServerPathToJavaSource(),className);
        OutputFile out = new OutputFile(fileName);
        
        out.println("package " + getServerJavaPackageName() + ";\n\n");
        out.println("import java.io.*;");
        out.println(getBasicImportLines());
        out.println("import kosoft.dbgen.util.*;");

        out.println("public class " + className + " extends ClientConnection {\n");
            
        out.println("    public String getObjectTypeName() { return \"" + getDbConnectionName() +"\"; }\n");
    
    
    // Constructor
    
          out.println("\n    public " + className + "() {");
          out.println("    }\n");
   //      out.println("\n    public String getTableName() { return CONST_TableName; }\n\n") >> FileName;
    //      out.println("\n    public String getSelectHeader() { return CONST_SelectHeader; }\n\n") >> FileName;
    
    
        out.println("\n  public static buildHashtable() {\n");

        for (int i=0; i < tableList.size(); i++) {
            ObjectData obj =  (ObjectData) tableList.get(i);
            String objName = obj.getTableName();
            out.println("    registerObjectType(" + objName + "ProxyServer.PROXY_Name, " + objName + "ProxyServer.class);\n");
            out.println("    registerObjectType(" + objName + "ProxyQueryServer.PROXY_Name, " + objName + "ProxyQueryServer.class);\n");
            }
        out.println("\n  }\n\n");        
        out.println("}\n\n");
    
    }

    
   class ObjectData {
       boolean isUpdateableObjectFlg = true;
       String tableName;
       String sqlName;
       String stmt;
       
       void setSQLName(String newSqlName) {
           sqlName = newSqlName;
           tableName = convertNameFromSQL(sqlName);
       }
       
       String getTableName() { return tableName; }
       String getTableSqlName() { return sqlName; }
       String getFactoryName() { return tableName + "Factory"; }
       String getStmtName() { return tableName + "Stmt"; }
       String getTableStructName() { return tableName + "Struct"; }
       
       // Variable name has a lowercase first letter.
       String getTableVariableName() {
           return tableName.substring(0,1).toLowerCase() + tableName.substring(1); 
       }
       String getFactoryVariableName() { return getTableVariableName() + "Factory"; }
       String getStmtVariableName() { return getTableVariableName() + "Stmt"; }
       
       
       String objectType;
       ArrayList fields = new ArrayList();
       HashMap fieldLookup = new HashMap();
       
       ArrayList queries = new ArrayList();

       FieldData newField() { return new FieldData(); }
       StmtData newStmt() { return new StmtData(); }

       FieldData getField(String name) {
           return (FieldData) fieldLookup.get(name);
       }
    
       void addField(FieldData field) {
           field.fieldNum = fields.size();
           fields.add(field);
           fieldLookup.put(field.name, field);
       }


       String getFactoryMethodPrototype() {
           return "public " + getTableName() + "Factory get" + getTableName() + "Factory()";
       }
       
        
        void printProxyObjectServer() {
            String className = getTableName() + "ProxyServer";
            String fileName = buildFileName(getServerPathToJavaSource(),className);
            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getServerJavaPackageName() + ";\n");
            out.println("import java.io.*;\n\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import kosoft.dbgen.client.gen.*;\n\n");
            out.println("import " + getClientJavaPackageName() + "." + getTableName() +  ";\n\n");
            out.println("import " + getInterfacePackageName() + "."  + getTableName() + "Stmt;\n\n");
        
            out.println("public class " + className + " extends ProxyRowBaseServer implements " + getTableName() + "FieldsInterface {\n\n");
            
            out.println("  public String getObjectTypeName() { return PROXY_QueryName; }\n");
        
        
        // Constructor
        
              out.println("\n  public " + className + "() {\n");
              out.println("    oRow = new " + fileName +"();\n");
              out.println("  }\n\n");
        
              out.println("\n  public " + className + "(" + getTableName() + " oNewRow) {\n");
              out.println("    oRow = oNewRow;\n");
              out.println("  }\n\n");
        
        //      out.println("\n    public String getTableName()() { return CONST_TableName; }\n\n");
        //      out.println("\n    public String getSelectHeader() { return CONST_SelectHeader; }\n\n");
              out.println("}\n\n");
        }
        
        
        //------------------------------------------------------------------
        
        
        void printProxyObjectQueryServer() {
            String className = getTableName() + "ProxyQueryServer";
            String fileName = getServerPathToJavaSource() +  File.separator +  ".java";
            
            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getServerJavaPackageName() + ";\n");
            out.println("import java.io.*;\n\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n");
            out.println("import kosoft.dbgen.server.*;\n");
            out.println("import " + getClientGenJavaPackageName() + "." + getTableName() + "FieldsInterface;\n\n");
            out.println("import " + getInterfacePackageName() + "." + getTableName() + ";\n\n");
            out.println("import " + getInterfacePackageName() + "." + getTableName() + "Stmt;\n\n");
            
            out.println("public class " + className + " extends ProxyQueryBaseServer implements " + getTableName() + "FieldsInterface {\n\n");
    
            out.println("  public String getObjectTypeName() { return PROXY_QueryName; }\n");
        
        
        // Constructor
        
            out.println("\n  public " + className + "() {\n");
            out.println("    oQuery = new " + getTableName() + "Query();\n");
            out.println("  }\n\n");
        
            out.println("\n  public ProxyServerObjectBase BuildNewRowObject(UpdateableRowBase oNewRow) {\n");
            out.println("      return new " + getTableName() + "ProxyBase((" + getTableName() + ")oNewRow);\n");
            out.println("  }\n\n");
        
        
        //      out.println("\n    public String GetTableName() { return CONST_TableName; }\n\n") >> FileName;
        //      out.println("\n    public String GetSelectHeader() { return CONST_SelectHeader; }\n\n") >> FileName;
        
            out.println("}\n\n");
        }
        
        //------------------------------------------------------------------
        
        void printProxyFieldsInterface() {
            String className = getTableName() + "FieldsInterface";
            String fileName = getClientPathToJavaGenSource() + File.separator + className + ".java";
            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getClientGenJavaPackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("interface " + className + " {");
        
        // define costant field id's
        
            out.println("\n    static final String PROXY_ObjectName = \"ProxyObject" + getTableName() + "\";\n");
            out.println("\n    static final String PROXY_QueryName = \"ProxyQuery" + getTableName() + "\";\n");
        //    out.println("\n    static final String CONST_TableName = \"" +  TableName + "\";\n");
        //    out.println("    static final String CONST_SelectHeader = \"select * from " + TableName + "\";\n", );
        
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldIdDefinition(out);
            }
        
            out.println("\n");
        
        
        // NumFields method
        
            out.println("\n    static final int FIELDS_Max = " + Integer.toString(fields.size()) + ";\n");
            out.println("\n}\n");
        
            out.close();
        }
        
        //------------------------------------------------------------------
        
        
        void printProxyUpdateableObject() {
            String className = getTableName();
            String fileName = getClientPathToJavaSource() + File.separator + className + ".java";
        
        
        // Check if file has already been generated
        
            File file = new File(fileName);
            
            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getClientJavaPackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import " + getClientGenJavaPackageName() + ".*;\n\n");
            out.println("public class " +  className + " extends " + className + "ProxyBase {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
        
        }
        
        
        //------------------------------------------------------------------
        
        void printProxyUpdateableObjectBase() {
            String className = getTableName() + "ProxyBase";
            String fileName = getClientPathToJavaGenSource() + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);

            out.println("package " + getClientGenJavaPackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import kosoft.dbgen.client.*;\n\n");
            out.println("public class " + className + " extends ProxyUpdatableRowBase implements " + getTableName() + "FieldsInterface {\n");
        
            out.println("\n  public String getObjectTypeName() { return PROXY_ObjectName; }\n");
        
        // constructor
        
            out.println("  public " + className + "() {\n");
            out.println("    connectToServerObject();\n");
            out.println("  }\n\n");
        
        // NumFields method
        
            out.println("\n    public int NumFields() { return FIELDS_Max; }\n");
        //	out.println("\n    public String GetTableName() { return CONST_TableName; }\n");
        //	out.println("    public String GetSelectHeader() { return CONST_SelectHeader; }\n");
        
        
            String structName = "fieldValues";
            String origStructName = "origFieldValues";
        
            out.println("    " + getTableStructName() + " " + structName + " = new " + getTableStructName() + "()"); 
            out.println("    " + getTableStructName() + " " + origStructName + " = new " + getTableStructName() + "()"); 

            String fieldPrefix = structName + ".";
            String origFieldPrefix = origStructName + ".";

        
        // define fields and original fields
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldDef(out);
            }
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printOrigFieldDef(out);
            }
        
        
        // Read from Source method
        
            out.println("\n    public void setFromObjectArray(Object[] oInData) {\n");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldSetFromObjectArray(out);
            }                
            out.println("\n        SetOriginalFields();\n");
            out.println("\n    }\n");
        
        
        // SetOriginalFields method
        
            out.println("\n    public void SetOriginalFields(){\n");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetOrigField(out, fieldPrefix, origFieldPrefix);
            }                
            
            out.println("\n        bNewDataFlg = false;\n");
            out.println("\n    }\n");
        
        
        // GetFieldAsObject method
        
            out.println("\n    public Object getFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, fieldPrefix);
            }
            out.println("        }\n");
            out.println("        return null;\n");
            out.println("    }\n");
        
        
        // GetOrigFieldAsObject method
        
            out.println("\n    public Object getOrigFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return null;\n");
            out.println("    }\n");
        
        // SetFieldFromObject method
        
            out.println("\n    public void SetFieldFromObject(int iFieldNum, Object obj) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetFieldFromObject(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("    }\n");
        
        
        
        // FieldType method
        
            out.println("\n    public int FieldType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        
        // FieldDataType method
        
            out.println("\n    public int FieldDataType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldDataType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        // FieldValueAsString method
        
            out.println("\n    public String FieldValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsString(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        // FieldOrignalValueAsString method
        
            out.println("\n    public String FieldOriginalValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetOrigFieldAsString(out);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        
        // FieldHasBeenChanged method
        
            out.println("\n    public boolean FieldHasBeenChanged(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldHasBeenChanged(out, fieldPrefix, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return false;\n");
            out.println("    }\n");
        
        
        // FieldName method
        
            out.println("\n    public String FieldName(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldName(out);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");

            out.println("\n}\n");        
        }
        
                
        void printProxyObjectQuery() {
            String className = getTableName() + "Query";
            String fileName = getClientPathToJavaSource() + File.separator + className + ".java";
        
        
        // Check if file has already been generated
        
            File file = new File(fileName);
            
            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getClientJavaPackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");

            out.println("import " + getClientGenJavaPackageName() + ".*;\n\n");
            out.println("public class " +  className + " extends " + className + "ProxyQueryBase {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
            
        }
        
        
        //------------------------------------------------------------------
        
        
        void printProxyObjectQueryBase() {
            String className = getTableName() + "ProxyQueryBase";
            String fileName = getClientPathToJavaGenSource() + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getClientGenJavaPackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import kosoft.dbgen.client.*;\n\n");
            
            out.println("import " + getClientJavaPackageName() + "." + getTableName() + ";\n\n");
            out.println("public class " + className + " extends ProxyQueryBase implements " + getTableName() + "FieldsInterface {\n");
                
            out.println("  protected " + getTableName() + "[] oObjectData;\n\n");        
            out.println("  public String getObjectTypeName() { return PROXY_QueryName; }\n");
            
            out.println("\n    public int NumFields() { return FIELDS_Max; }\n\n");
        
        // Constructor
        
            out.println("\n  public " + className + "() {\n");
              
            out.println("    oDelegatedResultSet = new " + getTableName() + "();\n");
            out.println("    ConnectToServerObject();\n");
            out.println("  }\n\n");
        //      out.println("\n    public String GetTableName() { return CONST_TableName; }\n\n") >> FileName;
        //      out.println("\n    public String GetSelectHeader() { return CONST_SelectHeader; }\n\n") >> FileName;

            String argPrefix = "find";
            
            // Print find methods 
            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                if (stmtData.magnitude == 1) {
                    String findProto = stmtData.getFindMethodPrototype(argPrefix) + " {";
                    out.println(findProto);    
                
        // body of find lookup using the select clause
        
                    out.println("    " + className + " query = new " + className + "();");
                    out.println("    query." + stmtData.getSelectMethodName() + "("
                                 + stmtData.getParameterList(argPrefix) + ");");
                            
                    out.println("    " + getTableName() + "[] results = oQuery.Fetch();\n");
                    out.println("    if (results.length == 0) return null;\n");
                    out.println("    return results[0];");
        
                    out.println("  }");
        
                }
            }

            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                String findProto = stmtData.getSelectMethodPrototype(argPrefix) + " {";
                out.println(findProto);    
                
        
        // Generate where clauses for all the database entries
                ArrayList args = stmtData.argNames;
                for (int j=0; j< args.size(); j++) {
                    String argName = (String) args.get(j);
                    out.println("    AddWhere(FIELD_" + argName + ", "+ argPrefix  + argName + ");");
                }
        
                out.println("  }");
                    
            }
        
        // Fetch method
        
            out.println("\n    public " + getTableName() + "[] Fetch() {");
        
            out.println("         FetchAll();");
        //      out.println("         return (%s[]) oRowData;\n",getTableName()) >> FileName;
            out.println("         return oObjectData;");
            out.println("    }");
        
        // AllocateArray method
        
            out.println("\n    public void AllocateArray() {");
        
            out.println("         oObjectData = new " + getTableName() + "[iNumRows];");
            out.println("         oRowData = oObjectData;");
            out.println("    }");
/*        
        // ReadRecord method
        //
        //	out.println("\n    public void ReadRecord() throws IOException {\n") >> FileName;
        //
        //      out.println("         %s oNew = new %s();\n",TableName, TableName) >> FileName;
        //	out.println("         oNew.setFromObjectArray(in);\n") >> FileName;
        //      out.println("         oRowData[iCurrentRow] = oNew;\n",TableName) >> FileName;
        //      out.println("    }\n") >> FileName;
        //
        
        // ReadRecord method
   */     
   
            out.println("\n    public void ReadRecord(Object[] oInData)  {");
        
            out.println("         " + getTableName() + " oNew = new " + getTableName() + "();");
            out.println("         oNew.setFromObjectArray(oInData);");
            out.println("         oRowData[iCurrentRow] = oNew;");
            out.println("    }\n");
        
        
            out.println("}\n");
            out.close();
         }
        
        //------------------------------------------------------------------
        
        
        void printObjectStmtInterface() {
            String className = getTableName() + "Stmt";
            String fileName = getPathToJavaInterface() +  File.separator +  className + ".java";
            
            File file = new File(fileName);
            
        // Check if file has already been generated

            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getInterfacePackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import " + getGenInterfacePackageName() + ".*;\n\n");
            out.println("public interface " +  className + " extends " + className + "Base {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
            out.close();
        }            

        
        void printObjectStmtInterfaceBase() {
            String className = getTableName() + "StmtBase";
            String fileName = getPathToJavaGenInterface() + File.separator + className + ".java";

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenInterfacePackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import " + getInterfacePackageName() + "." + getTableName() + ";\n");
            out.println("public interface " +  className + " extends DbGenStmtBase, " + getTableName() + "Fields {");
                            
            String argPrefix = "find";
/*            
            // Print find methods 
            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                if (stmtData.magnitude == 1) {
                    String findProto = stmtData.getFindMethodPrototype(argPrefix) + ";";
                    out.println(findProto);            
                }
        
            }
*/        
            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                String findProto = stmtData.getSelectMethodPrototype(argPrefix) + ";";
                out.println(findProto);    
            }
        
        // Fetch method
        
            //out.println("\n    public " + getTableName() + "[] Fetch();");
            //out.println("\n    public void AllocateArray();");
            //out.println("\n    public void ReadRecord(Object[] oInData);");        
            out.println("}\n");
            out.close();
         }
        
        void printObjectFactoryInterface() {
            String className = getTableName() + "Factory";
            String fileName = getPathToJavaInterface() +  File.separator +  className + ".java";
            
            File file = new File(fileName);
            
        // Check if file has already been generated

            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getInterfacePackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import " + getGenInterfacePackageName() + ".*;\n\n");
            out.println("public interface " +  className + " extends " + className + "Base {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
            
            out.close();
        }            

        
        void printObjectFactoryInterfaceBase() {
            String className = getTableName() + "FactoryBase";
            String fileName = getPathToJavaGenInterface() + File.separator + className + ".java";

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenInterfacePackageName() + ";\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n");
            out.println("import " + getInterfacePackageName() + "." + getTableName() + ";");
            out.println("import " + getInterfacePackageName() + "." + getTableName() + "Stmt;\n");
            out.println("public interface " +  className + " extends DbGenFactory, " + getTableName() + "Fields {");
                            
            String argPrefix = "find";
            
            // Print find methods 
            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                    String findProto = stmtData.getFindMethodPrototype(argPrefix) + ";";
                    out.println(findProto);
            }

            out.println("    public " + getStmtName() + " getStmt();\n");
            out.println("    public " + getTableName() + " newInstance();");

/*           
            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                String findProto = stmtData.getSelectMethodPrototype(argPrefix) + ";";
                out.println(findProto);    
            }
*/
            out.println("\n}\n");
            out.close();
            
         }

        
        void printObjectStmt(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getStmtName() + extension;
            String fileName = getSourcePathName(mode) + File.separator + className + ".java";
        
        
        // Check if file has already been generated
        
            File file = new File(fileName);
            
            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getPackageName(mode) + ";\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import " + getInterfacePackageName() + ".*;");
            out.println("import " + getGenPackageName(mode) + ".*;\n");
//            out.println("public class " +  className + " {\n");
            out.println("public class " +  className + " extends " + className + "Base {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
        
            out.close();
        }
        
        
        
        void printObjectStmtBase(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getStmtName() + extension + "Base";
            String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenPackageName(mode) + ";\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import kosoft.dbgen.rdb.*;\n");
            out.println("import " + getInterfacePackageName() + ".*;");
            out.println("import " + getPackageName(mode) + ".*;\n");
            out.println("public abstract class " +  className + " extends RdbStmtBase implements " + getStmtName() + " {");
        
//            out.println("\n    public int getNumFields() { return FIELDS_Max; }\n");
            
        // Constructor        
              out.println("    public " + className + "() {");
              out.println("    }\n");
        //      out.println("\n    public String GetTableName() { return CONST_TableName; }\n");

            String argPrefix = "find";
                    
            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                String findProto = stmtData.getSelectMethodPrototype(argPrefix) + " {";
                out.println(findProto);    
                
        
        // Generate where clauses for all the database entries
                ArrayList args = stmtData.argNames;
                for (int j=0; j < args.size(); j++) {
                    String argName = (String) args.get(j);
                    out.println("        addWhere(FIELD_" + argName + ", "+ argPrefix  + argName + ");");
                }
        
                out.println("    }");

                    
            }
                
            out.println("}\n");
            
            out.close();
         }

        
        void printObjectFactory(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getFactoryName() + extension;
            String fileName = getSourcePathName(mode) + File.separator + className + ".java";
        
        
        // Check if file has already been generated
        
            File file = new File(fileName);
            
            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getPackageName(mode) + ";\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n");
            out.println("import " + getInterfacePackageName() + ".*;\n");
            out.println("import " + getGenPackageName(mode) + ".*;\n");
//            out.println("public class " +  className + " {\n");
            out.println("public class " +  className + " extends " + className + "Base {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
            out.close();
        }
        
        
        
        void printObjectFactoryBase(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getFactoryName() + extension + "Base";
            String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenPackageName(mode) + ";\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import kosoft.dbgen.rdb.*;\n");
            out.println("import " + getInterfacePackageName() + ".*;");
            out.println("import " + getPackageName(mode) + ".*;\n");
            out.println("public abstract class " +  className + " extends RdbFactoryBase implements " + getFactoryName() + " {");
        
//            out.println("\n    public int getNumFields() { return FIELDS_Max; }");
            
        // Constructor        
              out.println("    public " + className + "() {");
              out.println("    }\n");
        //      out.println("\n    public String GetTableName() { return CONST_TableName; }\n");
        //      out.println("\n    public String GetSelectHeader() { return CONST_SelectHeader; }\n");
        
            String argPrefix = "find";
            String buildClassName = getStmtName() + extension;
            
            // Print find methods 
            for (int i=0; i < queries.size(); i++) {
                StmtData stmtData = (StmtData) queries.get(i);
                String findProto = stmtData.getFindMethodPrototype(argPrefix) + " {";
                out.println(findProto);    
        
        // body of find lookup using the select clause
                out.println("        " + buildClassName + " newStmt = new " + buildClassName + "();");
                //out.println("        " + getStmtName() + " stmt = new " + getStmtName() + extension + "();");
                out.println("        newStmt.setFactory(this);");
                out.println("        newStmt." + stmtData.getSelectMethodName() + "("
                                 + stmtData.getArgsList(argPrefix) + ");");

                if (stmtData.magnitude == 1) {
                    out.println("        Iterator iter = newStmt.execute().iterator();");
                    out.println("        if (!iter.hasNext()) {");
                    out.println("            return null;");
                    out.println("        }");
                    out.println("        return (" + getTableName() + ") iter.next();");
                }
                else {
                    out.println("        return newStmt.execute();");
                }
        
                out.println("    }");
        
            }
        
            out.println("    public " + getTableName() + "Stmt getStmt() { ");
            out.println("        " + buildClassName + " newStmt = new " + buildClassName + "();");
            out.println("        newStmt.setFactory(this);");
            out.println("        return newStmt;");
            out.println("     }");
            out.println("");

            out.println("    public RdbStmtBase  newStmt() {");
            out.println("        return (RdbStmtBase) getStmt();");
            out.println("    }");
            out.println("");
                
            
            out.println("    public " + getTableName() + " newInstance() {");
            buildClassName = getTableName() + extension;
            out.println("        " + buildClassName + " newObj = new " + buildClassName + "();");
            out.println("         newObj.setFactory(this);");
            out.println("         return newObj;");
            out.println("     }");
            out.println("");
            
            out.println("    public RdbRowBase  newRow() {");
            out.println("        return (RdbRowBase) newInstance();");
            out.println("    }");
            out.println("");

            out.println("}\n");
            out.close();
         }


         
        void printFieldsInterface() {
            String className = getTableName() + "Fields";
            String fileName = getPathToJavaGenInterface() + File.separator + className + ".java";

            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenInterfacePackageName() + ";\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("public interface " +  className + " {");
        
        // define costant field id's
        
            //out.println("\n    static final String CONST_TableName = \"" + getTableName() + "\";");
            //out.println("    static final String CONST_SelectHeader = \"select * from " + getTableName() + "\";");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldIdDefinition(out);
            }
            
        
            out.println("\n");
        
        
        // NumFields method
        
            out.println("\n    static final int FIELDS_Max = " + Integer.toString(fields.size()) + ";\n");
            out.println("\n}\n");
        
            out.close();
        }
        
        //------------------------------------------------------------------
        
        void printBaseViewObject(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getTableName() + extension + "Base";
            String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenPackageName(mode) + ";\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import kosoft.dbgen.client.*;\n\n");
            out.println("import " + getInterfacePackageName() + ".*;");
            out.println("public class " + className + " extends RowBase implements " + getTableName() + " {\n");
                
        // constructor
        
            out.println("  public " + className + "() {}");
        
        // NumFields method
        
            out.println("\n    public int getNumFields() { return FIELDS_Max; }\n");
        //	out.println("\n    public String GetTableName() { return CONST_TableName; }\n");
        //	out.println("    public String GetSelectHeader() { return CONST_SelectHeader; }\n");
        
        
        
        // define fields and original fields
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldDef(out);
            }
        
        
        // Read from Source method
        
            out.println("\n    public void setFromObjectArray(Object[] oInData) {\n");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldSetFromObjectArray(out);
            }                
            out.println("\n    }\n");
        
            out.close();
        }
        
        
        //------------------------------------------------------------------
        
        
        void printObject(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getTableName() +  extension;
            String fileName = getSourcePathName(mode) + File.separator + className + ".java";
            
        
        // Check if file has already been generated
        
            File file = new File(fileName);
            
            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getPackageName(mode) + ";\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import " + getInterfacePackageName() + ".*;");
            out.println("import " + getGenPackageName(mode) + "." + className + "Base;\n");
            out.println("public class " +  className + " extends " + className + "Base {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
        
            out.close();
        }
        

        void printObjectStruct(String mode) {
//        String extension = getExtensionForMode(mode);
            String extension = "";
            String className = getTableStructName() + extension;
            String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenPackageName(mode) + ";\n\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import " + getInterfacePackageName() + ".*;\n");
            out.println("public class " + className + " {\n");
        
        
        // constructor
        
            //out.println("  public " + className + "() {\n");
            //out.println("  }\n");
        
        // NumFields method
        
            //out.println("\n    public int getNumFields() { return FIELDS_Max; }\n");
            //out.println("\n    public String getTableName() { return CONST_TableName; }");
            //out.println("    public String getSelectHeader() { return CONST_SelectHeader; }");
        
        
        
        // define fields and original fields
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldDef(out);
            }
        
        // Read from Source method
        

            out.println("\n}\n");        

            out.close();
        }
        
        //------------------------------------------------------------------

        void printBaseUpdatableObject(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getTableName() + extension + "Base";
            String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenPackageName(mode) + ";\n\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import kosoft.dbgen.rdb.*;");
            out.println("import " + getInterfacePackageName() + ".*;\n");
            out.println("public abstract class " + className + " extends RdbUpdatableRowBase implements " + getTableName() + " {\n");
        
        
        // constructor
        
            out.println("    public " + className + "() {");
            out.println("    }\n");
        
        // NumFields method
            out.println("\n    static final String CONST_TableName = \"" + getTableSqlName() + "\";\n");
            out.println("    static final String CONST_SelectHeader = \"select * from " + getTableSqlName() + "\";");
        
            out.println("\n    public int getNumFields() { return FIELDS_Max; }");
            out.println("    public String getTableName() { return CONST_TableName; }");
            out.println("    public String getSelectHeader() { return CONST_SelectHeader; }");
        
            
            String structName = "fieldValues";
            String origStructName = "origFieldValues";
        
            out.println("");
            out.println("    " + getTableStructName() + " " + structName + " = new " + getTableStructName() + "();"); 
            out.println("    " + getTableStructName() + " " + origStructName + " = new " + getTableStructName() + "();"); 
            out.println("");

            String fieldPrefix = structName + ".";
            String origFieldPrefix = origStructName + ".";

            // define fields and original fields
        
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldGetMethod(out, fieldPrefix);
                field.printFieldSetMethod(out, fieldPrefix);
                out.println("");
                //field.printFieldDef(out);
           }       
        
           //   for (int i=0; i < fields.size(); i++) {
           //     FieldData field = (FieldData) fields.get(i);
           //     field.printOrigFieldDef(out);
           // }
        
        
        // Read from Source method
        
            out.println("\n    public void setFromObjectArray(Object[] oInData) {");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldSetFromObjectArray(out, fieldPrefix);
            }                
            out.println("\n        setOriginalFields();");
            out.println("    }\n");
        

        // Read from ResultSet
        
            out.println("\n    public void setFromResultSet(ResultSet row) throws SQLException {");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetFieldFromResultSet(out, fieldPrefix);
            }                
            out.println("\n        setOriginalFields();");
            out.println("    }\n");
        
        
        // SetOriginalFields method
        
            out.println("    public void setOriginalFields(){\n");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetOrigField(out, fieldPrefix, origFieldPrefix);
            }                
            
            out.println("\n        bNewDataFlg = false;");
            out.println("    }\n");
        
        
        // GetFieldAsObject method
        
            out.println("\n    public Object getFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, fieldPrefix);
            }
            out.println("        }");
            out.println("        return null;");
            out.println("    }\n");
        
        
        // GetOrigFieldAsObject method
        
            out.println("\n    public Object getOrigFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, origFieldPrefix);
            }

            out.println("        }");
            out.println("        return null;\n");
            out.println("    }\n");
        
        // SetFieldFromObject method
        
            out.println("\n    public void setFieldFromObject(int iFieldNum, Object obj) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetFieldFromObject(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("    }\n");
        
        
        
        // FieldType method
        
            out.println("\n    public int getFieldType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        
        // FieldDataType method
        
            out.println("\n    public int getFieldDataType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldDataType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        // FieldValueAsString method
        
            out.println("\n    public String getFieldValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsString(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        // FieldOrignalValueAsString method
        
            out.println("\n    public String getFieldOriginalValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsString(out, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        
        // FieldHasBeenChanged method
        
            out.println("\n    public boolean hasFieldChanged(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldHasBeenChanged(out, fieldPrefix, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return false;");
            out.println("    }\n");
        
        
        // FieldName method
        
            out.println("\n    public String getFieldName(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldName(out);
            }

            out.println("        }\n");
            out.println("        return new String();");
            out.println("    }\n");

        
        // FieldName method
        
            out.println("\n    public String getFieldSqlName(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldSqlName(out);
            }

            out.println("        }\n");
            out.println("        return new String();");
            out.println("    }\n");
            
            out.println("\n}\n");        

            out.close();
        }
        
        
        //------------------------------------------------------------------
        
        void printBaseUpdatableObjectXXX(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getTableName() + extension + "Base";
            String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenPackageName(mode) + ";\n\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import kosoft.dbgen.rdb.*;");
            out.println("import " + getInterfacePackageName() + ".*;\n");
            out.println("public abstract class " + className + " extends RdbUpdatableRowBase implements " + getTableName() + " {\n");
        
        
        // constructor
        
            out.println("    public " + className + "() {");
            out.println("    }\n");
        
        // NumFields method
            out.println("\n    static final String CONST_TableName = \"" + getTableSqlName() + "\";\n");
            out.println("    static final String CONST_SelectHeader = \"select * from " + getTableSqlName() + "\";");
        
            out.println("\n    public int getNumFields() { return FIELDS_Max; }");
            out.println("    public String getTableName() { return CONST_TableName; }");
            out.println("    public String getSelectHeader() { return CONST_SelectHeader; }");
        
            
            String structName = "fieldValues";
            String origStructName = "origFieldValues";
        
            out.println("");
            out.println("    " + getTableStructName() + " " + structName + " = new " + getTableStructName() + "();"); 
            out.println("    " + getTableStructName() + " " + origStructName + " = new " + getTableStructName() + "();"); 
            out.println("");

            String fieldPrefix = structName + ".";
            String origFieldPrefix = origStructName + ".";

            // define fields and original fields
        
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldGetMethod(out, fieldPrefix);
                field.printFieldSetMethod(out, fieldPrefix);
                out.println("");
                //field.printFieldDef(out);
           }       
        
           //   for (int i=0; i < fields.size(); i++) {
           //     FieldData field = (FieldData) fields.get(i);
           //     field.printOrigFieldDef(out);
           // }
        
        
        // Read from Source method
        
            out.println("\n    public void setFromObjectArray(Object[] oInData) {");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldSetFromObjectArray(out, fieldPrefix);
            }                
            out.println("\n        setOriginalFields();");
            out.println("    }\n");
        

        // Read from ResultSet
        
            out.println("\n    public void setFromResultSet(ResultSet row) throws SQLException {");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetFieldFromResultSet(out, fieldPrefix);
            }                
            out.println("\n        setOriginalFields();");
            out.println("    }\n");
        
        
        // SetOriginalFields method
        
            out.println("    public void setOriginalFields(){\n");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetOrigField(out, fieldPrefix, origFieldPrefix);
            }                
            
            out.println("\n        bNewDataFlg = false;");
            out.println("    }\n");
        
        
        // GetFieldAsObject method
        
            out.println("\n    public Object getFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, fieldPrefix);
            }
            out.println("        }");
            out.println("        return null;");
            out.println("    }\n");
        
        
        // GetOrigFieldAsObject method
        
            out.println("\n    public Object getOrigFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, origFieldPrefix);
            }

            out.println("        }");
            out.println("        return null;\n");
            out.println("    }\n");
        
        // SetFieldFromObject method
        
            out.println("\n    public void setFieldFromObject(int iFieldNum, Object obj) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetFieldFromObject(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("    }\n");
        
        
        
        // FieldType method
        
            out.println("\n    public int getFieldType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        
        // FieldDataType method
        
            out.println("\n    public int getFieldDataType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldDataType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        // FieldValueAsString method
        
            out.println("\n    public String getFieldValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsString(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        // FieldOrignalValueAsString method
        
            out.println("\n    public String getFieldOriginalValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsString(out, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        
        // FieldHasBeenChanged method
        
            out.println("\n    public boolean hasFieldChanged(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldHasBeenChanged(out, fieldPrefix, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return false;");
            out.println("    }\n");
        
        
        // FieldName method
        
            out.println("\n    public String getFieldName(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldName(out);
            }

            out.println("        }\n");
            out.println("        return new String();");
            out.println("    }\n");

        
        // FieldName method
        
            out.println("\n    public String getFieldSqlName(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldSqlName(out);
            }

            out.println("        }\n");
            out.println("        return new String();");
            out.println("    }\n");
            
            out.println("\n}\n");        

            out.close();
        }
        

        void printObjectBase(String mode) {
            if (isUpdateableObjectFlg) {
                printObjectStruct(mode);
                printBaseUpdatableObject(mode);
            }
            else {
                printBaseViewObject(mode);
            }
        }

        //------------------------------------------------------------------
        
        void printObjectInterface() {
            String className = getTableName();
            String fileName = getPathToJavaInterface() + File.separator + className + ".java";
            
        
        // Check if file has already been generated
        
            File file = new File(fileName);
            
            if (file.exists()) {
                System.out.println("file " + fileName + " already created");
                return;
            }

            OutputFile out = new OutputFile(fileName);
            
            out.println("package " + getInterfacePackageName() + ";\n\n");
            out.println("import java.io.*;\n");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;\n\n");
            out.println("import " + getGenInterfacePackageName() + ".*;\n\n");
            out.println("public interface " +  className + " extends " + className + "Base {\n");
            out.println("\n	// Generated Class which user can extend\n\n");
            out.println("\n}\n");
            out.close();
        
        }
        
        
        
        void printObjectInterfaceBase() {
            String className = getTableName() + "Base";
            String fileName = getPathToJavaGenInterface() + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenInterfacePackageName() + ";\n\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("public interface " + className + " extends DbGenRowBase, " + getTableName() + "Fields {\n");
//            out.println("public interface " + className + " extends " + getTableName() + "Fields {\n");
        
        // NumFields method
        

//            out.println("\n    public String getTableName() { return CONST_TableName; }");
//            out.println("    public String getSelectHeader() { return CONST_SelectHeader; }");
        /*
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldIdDefinition(out);
            }
            
        
        // NumFields method
        
            out.println("\n    static final int FIELDS_Max = " + Integer.toString(fields.size()) + ";");
            out.println("\n");
        */
                
        // define fields and original fields
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldPropertyPrototype(out);
                out.println("");
            }
        
/*        
            out.println("\n    public int getNumFields();");        
            out.println("\n    public Object getFieldAsObject(int iFieldNum);");
            out.println("\n    public void setFieldFromObject(int iFieldNum, Object obj);\n");

            out.println("\n    public Object getOrigFieldAsObject(int iFieldNum);");
            out.println("\n    public int getFieldType(int iFieldNum);");
            out.println("\n    public int getFieldDataType(int iFieldNum);");
            out.println("\n    public String getFieldValueAsString(int iFieldNum);");
            out.println("\n    public String getFieldOriginalValueAsString(int iFieldNum);");
            out.println("\n    public boolean hasFieldChanged(int iFieldNum);");
*/
            out.println("\n}\n");
            
            out.close();
        }
        
        //------------------------------------------------------------------

        void printBaseUpdatableObject(String mode) {
            String extension = getExtensionForMode(mode);
            String className = getTableName() + extension + "Base";
            String fileName = getGenSourcePathName(mode) + File.separator + className + ".java";
            
            OutputFile out = new OutputFile(fileName);

            out.println("package " + getGenPackageName(mode) + ";\n\n");
            out.println("import java.io.*;");
            out.println(getBasicImportLines());
            out.println("import kosoft.dbgen.util.*;");
            out.println("import kosoft.dbgen.rdb.*;");
            out.println("import " + getInterfacePackageName() + ".*;\n");
            out.println("public abstract class " + className + " extends RdbUpdatableRowBase implements " + getTableName() + " {\n");
        
        
        // constructor
        
            out.println("    public " + className + "() {");
            out.println("    }\n");
        
        // NumFields method
            out.println("\n    static final String CONST_TableName = \"" + getTableSqlName() + "\";\n");
            out.println("    static final String CONST_SelectHeader = \"select * from " + getTableSqlName() + "\";");
        
            out.println("\n    public int getNumFields() { return FIELDS_Max; }");
            out.println("    public String getTableName() { return CONST_TableName; }");
            out.println("    public String getSelectHeader() { return CONST_SelectHeader; }");
        
            
            String structName = "fieldValues";
            String origStructName = "origFieldValues";
        
            out.println("");
            out.println("    " + getTableStructName() + " " + structName + " = new " + getTableStructName() + "();"); 
            out.println("    " + getTableStructName() + " " + origStructName + " = new " + getTableStructName() + "();"); 
            out.println("");

            String fieldPrefix = structName + ".";
            String origFieldPrefix = origStructName + ".";

            // define fields and original fields
        
        
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldGetMethod(out, fieldPrefix);
                field.printFieldSetMethod(out, fieldPrefix);
                out.println("");
                //field.printFieldDef(out);
           }       
        
           //   for (int i=0; i < fields.size(); i++) {
           //     FieldData field = (FieldData) fields.get(i);
           //     field.printOrigFieldDef(out);
           // }
        
        
        // Read from Source method
        
            out.println("\n    public void setFromObjectArray(Object[] oInData) {");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printFieldSetFromObjectArray(out, fieldPrefix);
            }                
            out.println("\n        setOriginalFields();");
            out.println("    }\n");
        

        // Read from ResultSet
        
            out.println("\n    public void setFromResultSet(ResultSet row) throws SQLException {");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetFieldFromResultSet(out, fieldPrefix);
            }                
            out.println("\n        setOriginalFields();");
            out.println("    }\n");
        
        
        // SetOriginalFields method
        
            out.println("    public void setOriginalFields(){\n");

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetOrigField(out, fieldPrefix, origFieldPrefix);
            }                
            
            out.println("\n        bNewDataFlg = false;");
            out.println("    }\n");
        
        
        // GetFieldAsObject method
        
            out.println("\n    public Object getFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, fieldPrefix);
            }
            out.println("        }");
            out.println("        return null;");
            out.println("    }\n");
        
        
        // GetOrigFieldAsObject method
        
            out.println("\n    public Object getOrigFieldAsObject(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsObject(out, origFieldPrefix);
            }

            out.println("        }");
            out.println("        return null;\n");
            out.println("    }\n");
        
        // SetFieldFromObject method
        
            out.println("\n    public void setFieldFromObject(int iFieldNum, Object obj) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printSetFieldFromObject(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("    }\n");
        
        
        
        // FieldType method
        
            out.println("\n    public int getFieldType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        
        // FieldDataType method
        
            out.println("\n    public int getFieldDataType(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldDataType(out);
            }

            out.println("        }\n");
            out.println("        return -1;\n");
            out.println("    }\n");
        
        
        // FieldValueAsString method
        
            out.println("\n    public String getFieldValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsString(out, fieldPrefix);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        // FieldOrignalValueAsString method
        
            out.println("\n    public String getFieldOriginalValueAsString(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldAsString(out, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return new String();\n");
            out.println("    }\n");
        
        
        // FieldHasBeenChanged method
        
            out.println("\n    public boolean hasFieldChanged(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldHasBeenChanged(out, fieldPrefix, origFieldPrefix);
            }

            out.println("        }\n");
            out.println("        return false;");
            out.println("    }\n");
        
        
        // FieldName method
        
            out.println("\n    public String getFieldName(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldName(out);
            }

            out.println("        }\n");
            out.println("        return new String();");
            out.println("    }\n");

        
        // FieldName method
        
            out.println("\n    public String getFieldSqlName(int iFieldNum) {\n");
            out.println("         switch (iFieldNum) {");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                field.printGetFieldSqlName(out);
            }

            out.println("        }\n");
            out.println("        return new String();");
            out.println("    }\n");
            
            out.println("\n}\n");        

            out.close();
        }
        
        
        //------------------------------------------------------------------
        
        void printPythonObjectBase(OutputFile out) {
            String className = getTableName() + "Base";
            
            //out.println("package " + getGenPackageName(mode) + ";\n\n");
            out.println("class " + className + "(ResultSet):\n");
        // constructor

            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                out.println("    " + field.name + "= None");
                //field.printFieldDef(out);
           }       
        
        // NumFields method
            out.println("    numColumns = " + Integer.toString(fields.size()));
            out.println("    tableName = '" + getTableSqlName() + "'");
        
            out.println("    def getNumColumns(self):");
            out.println("        return self.numColumns;");
            out.println("");

            out.println("    def getTableName(self):");
            out.println("        return self.tableName;");
            out.println("");
            
            out.println("    def getNew(self):");
            out.println("        return " + className + "()");
            out.println("");

        // SQLFieldName method
        
            out.println("\n    def getSQLFieldName(self, i) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                out.println("        if i == " + Integer.toString(i) + ":");
                out.println("            return '" + field.sqlName + "'");
            }
            out.println("");
            

        // getField
        
            out.println("\n    def getField(self, i) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                out.println("        if i == " + i + ":");
                out.println("            return self." + field.name);
            }
            out.println("");


        // setField
        
            out.println("\n    def setField(self, i, newValue) {\n");
            for (int i=0; i < fields.size(); i++) {
                FieldData field = (FieldData) fields.get(i);
                out.println("        if i == " + i + ":");
                out.println("            self." + field.name + " = newValue");
            }
            out.println("");
            
        }
        
        
       class FieldData {
           String sqlName;
           String name;
           String dataType;
           int fieldNum;
           boolean primaryKey = false;
    
           void setSQLName(String newSQLName) {
               sqlName = newSQLName;
               name = convertNameFromSQL(sqlName);
           }
           
           String getJavaType() {
               SqlTypeMapEntry entry = (SqlTypeMapEntry) sqlToJavaTypeMap.get(dataType);
               if (entry == null) {
                   System.out.println("Unknown type " + dataType);
                   return null;
               }
               return entry.javaType;
//               return getJavaType(dataType);               
           }

           void printFieldDef(OutputFile out) {
               out.println("    public " +  getJavaType() + " " + name + ";");
           }       
           void printOrigFieldDef(OutputFile out) {
               out.println("    public " +  getJavaType() + " Orig" + name + ";");
           }
           void printFieldSetFromObjectArray(OutputFile out) {
               String subField = "oInData[" + fieldNum + "]";
               out.println("        " + name + " = " + getConversionToJava(dataType, subField) + ";");
           }
           void printFieldSetFromObjectArray(OutputFile out, String fieldPrefix) {
               String subField = "oInData[" + fieldNum + "]";
               out.println("        " + fieldPrefix + name + " = " + getConversionToJava(dataType, subField) + ";");
           }
    
           void printSetOrigField(OutputFile out, String fieldPrefix, String origFieldPrefix) {
               out.println("        " + origFieldPrefix + name + " = " + fieldPrefix + name + ";");
           }
           
           void printGetFieldAsObject(OutputFile out, String fieldPrefix) {
               out.println("         case FIELD_" + name + ":");
               String fieldRef = fieldPrefix + name; 
               out.println("             return " + getConversionToObject(dataType, fieldRef) + ";");
               //	    out.println("             break;\n") >> FileName;
           }
           void printGetFieldAsString(OutputFile out, String fieldPrefix) {
               out.println("         case FIELD_" + name + ":");
               String retValue = getConvToString(fieldPrefix + name, dataType);
               out.println("             return " + retValue + ";");
               //	    out.println("             break;\n") >> FileName;
           }
           
           void printSetFieldFromObject(OutputFile out, String fieldPrefix) {
               out.println("         case FIELD_" + name + ":");
               out.println("             " + fieldPrefix + name + " = " + getConversionToJava(dataType, "obj") + ";");
               out.println("             break;");
           }

           void printSetFieldFromResultSet(OutputFile out, String fieldPrefix) {
               out.println("         " + fieldPrefix + name + " = row." + getConvFromResultSet(fieldNum + 1, dataType) + ";");
           }
           
           void printGetFieldType(OutputFile out) {
               String fieldFType = "FTYPE_Data";
               if (name.equals("RowID")) {
                   fieldFType = "FTYPE_RowID";
               }
               else if (primaryKey) {
                   fieldFType = "FTYPE_PrimaryKey";
               }
               
               out.println("         case FIELD_" + name + ":");
               out.println("             return " + fieldFType + ";");
               //	    out.println("             break;\n") >> FileName;
           }
           void printGetFieldDataType(OutputFile out) {
               out.println("         case FIELD_" + name + ":");
               String dataString = getJavaType();
               if (dataString.equals("byte[]")) {
                   dataString = "ByteArray";
               }
               out.println("             return FDATATYPE_" + dataString + ";");
               //	    out.println("             break;\n") >> FileName;
           }
           
           //void printGetOrigFieldAsObject(OutputFile out) {
           //    out.println("         case FIELD_" + name + ":\n");
           //    String subName = "Orig" + name;
           //    out.println("             return " + getConversionToObject(dataType, subName) + ";\n");
           //    //	    out.println("             break;\n") >> FileName;
           //}

           void printGetOrigFieldAsString(OutputFile out) {
               out.println("         case FIELD_" + name + ":");
               String retValue = getConvToString("Orig" + name, dataType);
               out.println("             return " + retValue + ";");
               //	    out.println("             break;\n") >> FileName;
           }
           void printGetFieldHasBeenChanged(OutputFile out, String fieldPrefix, String origPrefix) {
               String javaType = getJavaType();
               out.println("         case FIELD_" + name + ":");
               if (javaType.equals("String") || javaType.equals("DateTime") ) {
                   out.println("             return (!areFieldsEqual(" + fieldPrefix + name + "," + origPrefix + name + "));");
            // 	         out.println("             return (!Orig%s.equals(%s));\n",Name,Name) >> FileName;
               } else {
                   out.println("             return (" + fieldPrefix + name + " != " + origPrefix + name +");");
               }
           }
           void printGetFieldName(OutputFile out) {
               out.println("         case FIELD_" + name + ":");
               out.println("             return \"" + name + "\";");
           }
           void printGetFieldSqlName(OutputFile out) {
               out.println("         case FIELD_" + name + ":");
               out.println("             return \"" + sqlName.toLowerCase() + "\";");
           }

           void printFieldPropertyPrototype(OutputFile out) {
               out.println("    " + getFieldGetPrototype() + ";");
               out.println("    " + getFieldSetPrototype() + ";");               
           }       

           String getFieldGetPrototype() {
               return "public " +  getJavaType() + " get" + name + "()";
           }
           String getFieldSetPrototype() {
               return "public void set" + name + "(" + getJavaType() + " newValue)";
           }

                      
           void printFieldGetMethod(OutputFile out, String fieldPrefix) {
               out.println("    " + getFieldGetPrototype() + " { return " + fieldPrefix + name + "; }" );
           }
           
           void printFieldSetMethod(OutputFile out, String fieldPrefix) {
               out.println("    " + getFieldSetPrototype() + " {");
               out.println("        changingColumn(FIELD_" + name + ");");
               out.println("        " + fieldPrefix + name + " = newValue;");
               out.println("    }");
           }           
           void printFieldIdDefinition(OutputFile out) {
               out.println("    static final int FIELD_" + name + " = " + Integer.toString(fieldNum) + ";");
           }
     
            String buildGetFieldCall() {
               return " get" + name + "()";
           }       

            String buildSetFieldCall(String value) {
               return " set" + name + "(" + value + ")";
           }       

       }
       
       class StmtData {
           String name;
           int magnitude;
           ArrayList argNames = new ArrayList();
    
           String getFindMethodName() {
               String methodName = "get";           
               if (!name.equals("Find")) {
                   methodName = "getVia" + name;
               }
               return methodName;
           }
    
           String getSelectMethodName() {
               String methodName = "select";           
               if (!name.equals("Find")) {
                   methodName = "selectVia" + name;
               }
               return methodName;
           }
           
           String getParameterDefList() {
               return getParameterDefList("");
           }
           String getParameterDefList(String namePrefix) {
               String result = "";
               for (int i =0; i < argNames.size(); i++) {
                   String fieldName = (String) argNames.get(i);
                   if (i != 0) {
                       result += ", ";
                   }
                   FieldData field = getField(fieldName);
                   String fieldType = field.getJavaType();
                   result += fieldType + " " + namePrefix + fieldName;
               }
               return result;
           }
    
           String getParameterList() {
               return getParameterDefList("");
           }
           
           String getParameterList(String namePrefix) {
               String result = "";
               for (int i =0; i < argNames.size(); i++) {
                   String argName = (String) argNames.get(i);
                   FieldData field = getField(argName);
                   if (i != 0) {
                       result += ", ";
                   }
                   String fieldType = field.getJavaType();
                   result += fieldType + " " + namePrefix + argName;
               }
               return result;
           }

           String getArgsList() {
               return getArgsList("");
           }
           
           // List of arguments without the type definitions.
           String getArgsList(String namePrefix) {
               String result = "";
               for (int i =0; i < argNames.size(); i++) {
                   String argName = (String) argNames.get(i);
                   FieldData field = getField(argName);
                   if (i != 0) {
                       result += ", ";
                   }
                   result += namePrefix + argName;
               }
               return result;
           }
           
           String  getFindMethodPrototype(String prefix) {
               String result = "    public ";
               if (magnitude == 1) {
                   result += getTableName() + " " + getFindMethodName(); 
               }
               else {
                   result += "Collection " + getSelectMethodName(); 
               }
               result += "(" + getParameterDefList(prefix) + ")";

               return result;               
           }
           
           String  getSelectMethodPrototype(String prefix) {
               return "    public void " + getSelectMethodName() 
                             + "(" + getParameterDefList(prefix) + ")";
               
           }
           
       }
        
    }


           // Convert format from SQLName to java Name style 
        
    String convertNameFromSQL(String sqlName) {
       String newName = "";
       StringTokenizer tk = new StringTokenizer(sqlName, "_");
       while (tk.hasMoreTokens()) {
           String word = tk.nextToken();
           if (word.length() > 1) {
               String newWord = word.charAt(0) + word.substring(1).toLowerCase();
               newName += newWord;
           }
           else {
               newName += word;
           }
       }
       return newName;
    }

    
    String getJavaType(String sqlTypeName) {
       SqlTypeMapEntry entry = (SqlTypeMapEntry) sqlToJavaTypeMap.get(sqlTypeName);
       return entry.javaType;
    }

    int getSqlType(String sqlTypeName) {
       SqlTypeMapEntry entry = (SqlTypeMapEntry) sqlToJavaTypeMap.get(sqlTypeName);
       return entry.sqlType;
    }

    
    static HashMap sqlToJavaTypeMap = new HashMap();
    
    
    static void addItem(int sqlType, String sqlName, String javaType, String fromResultSet,  boolean isObj) {
        SqlTypeMapEntry entry = new SqlTypeMapEntry(sqlType, sqlName, javaType, fromResultSet, isObj);
        sqlToJavaTypeMap.put(sqlName, entry);
    }

    /*
    static {    
        sqlToJavaTypeMap.put("Number","int");
        sqlToJavaTypeMap.put("String", "String");
        sqlToJavaTypeMap.put("Boolean", "boolean");
        sqlToJavaTypeMap.put("DateTime", "Date");
    
   //     DataSource="TisBank_Local";
    
    }
    */
    
    static {
        addItem(Types.DATE, "DATE", "Date", "Date", true); 
        addItem(Types.TIMESTAMP, "TIMESTAMP", "Timestamp", "Timestamp", true); 
        addItem(Types.DOUBLE, "DOUBLE", "double", "Double", false); 
        addItem(Types.FLOAT, "FLOAT", "float", "Float",  false); 
        addItem(Types.INTEGER, "INTEGER", "int", "Int",  false); 
        addItem(Types.NUMERIC, "NUMERIC", "int", "Int", false); 
        addItem(Types.LONGVARBINARY, "LONGVARBINARY", "byte[]", "Bytes",  true); 
        addItem(Types.LONGVARCHAR, "LONGVARCHAR", "String", "String", true); 
        addItem(Types.VARCHAR, "VARCHAR", "String", "String", true); 
        addItem(Types.CHAR, "CHAR", "String", "String", false); 
/*
            addItem(Types.ARRAY, "ARRAY", null, true);  
            addItem(Types.BIGINT, "BIGINT", "BigInteger", true); 
            addItem(Types.BINARY, "BINARY", null, true); 
            addItem(Types.BIT, "BIT", null, true); 
            addItem(Types.BLOB, "BLOB", null, true); 
            addItem(Types.BOOLEAN, "BOOLEAN", "boolean", false); 
            addItem(Types.CLOB, "CLOB", null, true); 
            addItem(Types.DATALINK, "DATALINK", null, true); 
            addItem(Types.DECIMAL, "DECIMAL", "long", false); 
            addItem(Types.DISTINCT, "DISTINCT", null); 
            addItem(Types.JAVA_OBJECT, "JAVA_OBJECT", null, true); 
            addItem(Types.NULL, "NULL", null, false); 
            addItem(Types.OTHER, "OTHER", null); 
            addItem(Types.REAL, "REAL", null); 
            addItem(Types.REF, "REF", null); 
            addItem(Types.SMALLINT, "SMALLINT", null); 
            addItem(Types.STRUCT, "STRUCT", null); 
            addItem(Types.TIME, "TIME", null); 
            addItem(Types.TIMESTAMP, "TIMESTAMP", null); 
            addItem(Types.TINYINT, "TINYINT", null); 
            addItem(Types.VARBINARY, "VARBINARY", null); 
            */
    }

    String getSqlTypeNameFromType(int inType) {
//    String convertSQLTypeToKnownType(int inType) {
            switch (inType) {
                case Types.ARRAY: 
                    return "ARRAY";
                case Types.BIGINT: 
                    return "BIGINT";
                case Types.BINARY: 
                    return "BINARY";
                case Types.BIT: 
                    return "BIT";
                case Types.BLOB: 
                    return "BLOB";
                case Types.BOOLEAN: 
                    return "BOOLEAN";
                case Types.CHAR: 
                    return "CHAR";
                case Types.CLOB: 
                    return "CLOB";
                case Types.DATALINK: 
                    return "DATALINK";
                case Types.DATE: 
                    return "DATE";
                case Types.DECIMAL: 
                    return "DECIMAL";
                case Types.DISTINCT: 
                    return "DISTINCT";
                case Types.DOUBLE: 
                    return "DOUBLE";
                case Types.FLOAT: 
                    return "FLOAT";
                case Types.INTEGER: 
                    return "INTEGER";
                case Types.JAVA_OBJECT: 
                    return "JAVA_OBJECT";
                case Types.LONGVARBINARY: 
                    return "LONGVARBINARY";
                case Types.LONGVARCHAR: 
                    return "LONGVARCHAR";
                case Types.NULL: 
                    return "NULL";
                case Types.NUMERIC: 
                    return "NUMERIC";
                case Types.OTHER: 
                    return "OTHER";
                case Types.REAL: 
                    return "REAL";
                case Types.REF: 
                    return "REF";
                case Types.SMALLINT: 
                    return "SMALLINT";
                case Types.STRUCT: 
                    return "STRUCT";
                case Types.TIME: 
                    return "TIME";
                case Types.TIMESTAMP: 
                    return "TIMESTAMP";
                case Types.TINYINT: 
                    return "TINYINT";
                case Types.VARBINARY: 
                    return "VARBINARY";
                case Types.VARCHAR: 
                    return "VARCHAR";
                default:
                    return "Unknown";
            }
        }

/*        
    String convertSQLTypeToKnownType(String InType) {
        if (InType.equals("LONG" ))          return "Number";
        else if (InType.equals("COUNTER"))  return "Number";
        else if (InType.equals("BYTE"))  return "Number";
        else if (InType.equals("TEXT"))      return "String";
        else if (InType.equals("LONGTEXT"))  return "String";
        else if (InType.equals("DATETIME"))  return "DateTime";
        else if (InType.equals("BIT"))       return "Boolean";
        else {
            System.out.println("Unknown type : " + InType);
            return InType;
        }
    
    }
*/
    
    String getSQLField(String Name, String sqlTypeName) {
        
        String result;
        String FType = getJavaType(sqlTypeName);
        if (FType.equals("String")) {
            result = "'" + Name + "'";
        } else if (FType.equals("Date")) {
            result = "%d " + Name;
        } else {
          result = Name;
        }
        return result;
    }
    
    String getConversionToJava(String sqlTypeName, String SubField) {
        String javaType = getJavaType(sqlTypeName);
        if (javaType.equals("int")) {
            return "((Integer)" + SubField + ").intValue()";
        }
        else if (javaType.equals("String")) {
            return "(String) " + SubField;
        }
        else if (javaType.equals("float")) {
            return "((Float)" + SubField + ").floatValue()";
        }
        else if (javaType.equals("double")) {
            return "((Double)" + SubField + ").doubleValue()";
        }
        else if (javaType.equals("long")) {
            return "((Long)" + SubField + ").longValue()";
        }
        else if (javaType.equals("char")) {
            return "((Char)" + SubField + ").charValue()";
        }
        else if (javaType.equals("byte[]")) {
            return "(byte[])" + SubField;
        }
        else if (javaType.equals("boolean")) {
            return "((Boolean)" + SubField + ").booleanValue()";
        }
        else if (javaType.equals("Date")) {
            return "(Date) " + SubField;
        }
        else {
            return "(" + javaType + ") " + SubField;
        }
    
    }
    
    String getConversionToObject(String sqlTypeName, String SubField) {
        String javaType = getJavaType(sqlTypeName);
        if (javaType.equals("int")) {
            return "new Integer(" + SubField + ")";
        }
        else if (javaType.equals("String")) {
            return SubField;
        }
        else if (javaType.equals("float")) {
            return "new Float(" + SubField + ")";
        }
        else if (javaType.equals("double")) {
            return "new Double(" + SubField + ")";
        }
        else if (javaType.equals("long")) {
            return "new Long(" + SubField + ")";
        }
        else if (javaType.equals("byte[]")) {  
            return SubField;
        }
        else if (javaType.equals("boolean")) {
            return "new Boolean(" + SubField + ")";
        }
        else if (javaType.equals("char")) {
            return "new Char(" + SubField + ")";
        }
        else if (javaType.equals("Date")) {
            return SubField;
        }
        else {
            return SubField;
        }
    
    }
    
    String getConvToString(String Name,String sqlTypeName) {
        String javaType = getJavaType(sqlTypeName);
        String result;
        if (javaType.equals("String")) {
            result = Name;
        } 
        else if (javaType.equals("Date")) {
            result = "DateToSQLString(" + Name + ")";
    //	result = sout.println("%s.toString()", Name);
        } 
        else if (javaType.equals("Timestamp")) {
            result = "DateTimeToSQLString(" + Name + ")";
    //	result = sout.println("%s.toString()", Name);
        } 
        else if (javaType.equals("int")) {
            result = "Integer.toString(" + Name + ")";
        } 
        else if (javaType.equals("boolean")) {
            result = "BooleanToSQLString(" + Name + ")";
    //	result = sout.println("new Boolean(%s).toString()", Name);
        } 
        else if (javaType.equals("float")) {
            return "Float.toString(" + Name + ")";
        }
        else if (javaType.equals("double")) {
            return "Double.toString(" + Name + ")";
        }
        else if (javaType.equals("long")) {
            return "Long.toString(" + Name + ")";
        }
        else if (javaType.equals("char")) {
            return "Char.toString(" + Name + ")";
        }
        else if (javaType.equals("byte[]")) {  
            return "Base64.encode(" + Name + ")";
        } 
        else {
          result = javaType + ".toString(" + Name + ")";
        }
        return result;
    }
    
    String getConvFromResultSet(int columnNum, String sqlTypeName) {
       SqlTypeMapEntry entry = (SqlTypeMapEntry) sqlToJavaTypeMap.get(sqlTypeName);
       String methodName = "get" + entry.fromResultSet;
       return  methodName + "(" + Integer.toString(columnNum) + ")";
    }
    
   
   static final String TAG_Table = "Table";
   static final String TAG_Stmt = "stmt";
   static final String TAG_Objtype = "objtype";
   static final String TAG_Field = "field";
   static final String TAG_PrimaryQuery = "primaryquery";
   static final String TAG_Query = "query";
   static final String TAG_EndTable = "end_table";
   
   
   
   
   public void run(String fileName ) throws Exception {
      ObjectData obj = null;

      readParameters();
      setupDirs();
      
      BufferedReader in = new BufferedReader(new FileReader(fileName));
      
      String line = in.readLine();
      while(line != null) {
          if (line.startsWith(TAG_Table)) {
              obj = new ObjectData();
              String tableName = line.substring(TAG_Table.length()).trim();
              obj.setSQLName(tableName);
          }
          else if (line.startsWith(TAG_Stmt)) {
              obj.stmt = line.substring(TAG_Stmt.length()).trim();
          }
          else if (line.startsWith(TAG_Objtype)) {
              obj.objectType = line.substring(TAG_Objtype.length()).trim();
          }
          else if (line.startsWith(TAG_Field)) {
              ObjectData.FieldData field = obj.newField();
              String dataLine = line.substring(TAG_Field.length()).trim();
              StringTokenizer tk = new StringTokenizer(dataLine);              
              String fieldName = tk.nextToken();
              String fieldTypeName = tk.nextToken();
//              int dataType = Integer.parseInt(tk.nextToken());
              field.setSQLName(fieldName);
//              System.out.println(" " + dataLine + " n:" + fieldName + " t:" + fieldType); 
              field.dataType = fieldTypeName;
//              field.dataType = convertSQLTypeToKnownType(fieldType);
              obj.addField(field);
          }
          else if (line.startsWith(TAG_PrimaryQuery)) {
              ObjectData.StmtData stmt = obj.newStmt();
              stmt.name = "Find";
              stmt.magnitude = 1;
              StringTokenizer tk = new StringTokenizer(line.substring(TAG_PrimaryQuery.length()).trim());              
              while (tk.hasMoreTokens())   {
                  String fieldName = convertNameFromSQL(tk.nextToken());
                  stmt.argNames.add(fieldName);
                  ObjectData.FieldData fieldData = obj.getField(fieldName);
                  fieldData.primaryKey = true;
              }
              obj.queries.add(stmt);
          }
          else if (line.startsWith(TAG_Query)) {
              ObjectData.StmtData stmt = obj.newStmt();
              StringTokenizer tk = new StringTokenizer(line.substring(TAG_Query.length()).trim());              
              stmt.name = convertNameFromSQL(tk.nextToken());
              String mag = tk.nextToken();
              if (mag.equals("1")) {
                  stmt.magnitude = 1;
              }
              else {
                  stmt.magnitude = 2;
              }
              while (tk.hasMoreTokens())   {
                  String fieldName = convertNameFromSQL(tk.nextToken());
                  stmt.argNames.add(fieldName);
              }
              obj.queries.add(stmt);
          }
          else if (line.startsWith(TAG_EndTable)) {
              tableList.add(obj);
              
/*              
              
              obj.printProxyObjectQueryBase();
              obj.printProxyObjectQuery();
              obj.printProxyUpdateableObjectBase();
              obj.printProxyUpdateableObject();
              obj.printProxyFieldsInterface();
              
              obj.printProxyObjectServer();
              obj.printProxyObjectQueryServer();
*/
            

          }
          else {
              // Skip line not required.
          }
          
          line = in.readLine();
      }
      buildInterfaces();
//      buildDbClientConnection();
      buildRdbImpl();
      buildPython();
   }
   
   public static void main(String[] argv) throws Exception {

      BuildClassesV2 obj = new BuildClassesV2();
      obj.run(argv[0]);
   }
   
   
}


class SqlTypeMapEntry {
    int sqlType;
    String sqlName;
    String javaType;
    String fromResultSet;
    boolean isObject = false;
    
    SqlTypeMapEntry(int newType, String name, String javaClass, String newFromResultSet, boolean newIsObject) {
        sqlType = newType; 
        sqlName = name; 
        javaType = javaClass; 
        fromResultSet = newFromResultSet;
        isObject = newIsObject;
    }
}


class RuntimeProperties {
    Properties runParameters = new Properties();
    String propertyFileName = "jdbcgen.properties";

    void setFileName(String newName) { propertyFileName = newName; }
    
    void load() throws Exception {
        File loadFile = new File(propertyFileName);
        if (!loadFile.exists()) {
            throw new Exception("Missing required property file : " + propertyFileName);
        }
       FileInputStream input = new FileInputStream(loadFile);
       runParameters.load(input);       
    }

    String getProperty(String key) throws Exception {
        String value = getProperty(key, null);
        if (value == null) {
            throw new Exception("Missing required property " + key + " from " + propertyFileName);
        }
        return value;
    }
    String getProperty(String key, String defaultValue) {
        String value = runParameters.getProperty(key);
        if (value == null) {
            value = defaultValue;
        }
        return value;
    }

   double getPropertyAsDouble(String key) throws Exception {
       String valueString = getProperty(key);
       double value = -1.0;
       try {
           value = Double.parseDouble(valueString);
       }
       catch (NumberFormatException  e) {
            throw new Exception("Unable to parse number from " + value + "for key " + key + " in " + propertyFileName);
       }
       return value;
    }

   double getPropertyAsDouble(String key, double defaultValue) {
       double value = defaultValue;

       String valueString = getProperty(key, null);
       if (valueString != null) {
           try {
               value = Double.parseDouble(valueString);
           }
           catch (NumberFormatException  e) {}
       }
       return value;
    }

   boolean getPropertyAsBoolean(String key, boolean defaultValue) {
       boolean value = defaultValue;

       String valueString = getProperty(key, null);
       if (valueString != null) {
           if (valueString.toLowerCase().startsWith("t")) {
               value = true;
           }
           else if (valueString.toLowerCase().startsWith("f")) {
               value = false;
           }               
       }
       return value;
    }
}

